/* 

  Literal: 
     A literal is a variable with a sign. A variable is represented
     by a positive integer.
     One bit is used to denote the sign of a literal: 
              0 == positive, 1 == negative.

  Clause:
     Unit clauses are stored in Value[] or L5[x]->value.
     Binary clauses are stored in Bcs[].
     Other clauses are stored on tapes:
     A list of literals precedenced by its size (in negated form), head and
        tail indices.
     Two literals indicated by head/tail are head/tail literals. 
     Each literal is represented by a 32 bit integer, with one bit 
	representing its sign.

  Tape: 
     All non-unit, non-binary clauses are saved in an integerd array called Tape.
     An element in a tape can be (i) a literal, (ii) size of a clause,
        (iii) head/tail indices, (iv) zero, unused space, (v) values of quasi-clause.
 */

/***************************************************
 *
 *  TYPE Declaration
 *
 ***************************************************/

#define TAPE            int*
#define lit0()          4
#define lit1()          5
#define lit2()          6
#define lit3()          7

/***************************************************
 *
 *  MACRO Declaration
 *
 ***************************************************/

/* Macros for literals */
#define set_lit(p,ks) *(p) = ks

/* Macros for the special head/tail literals */
#define is_head(p) (*(p-1)<0)
#define get_head(cl) cl[cl[1]]
#define get_tail(cl) cl[cl[2]]
#define set_head_tail(p,i) *(p) = i
#define undeleted_head_tail(p) *p 
#define is_head_tail(cl,i) (i == cl[1] || i == cl[2])
#define is_literal(p) (*(p) > 0)
#define not_literal(p) (*(p) <= 0)
#define age_of_cl(cl) cl[3]
#define set_age_of_cl(cl,v) cl[3]=v
#define undeleted_clause(cl) cl[1] 
#define is_normal(cl) (cl[1] > 0) 

/* Macros for clauses */
/* acutual size plus lit0() */
#define set_clause_size(p,size) *p = -(size+lit0())
#define size_of_clause(p) (-*p)  
#define set_clause_address(p) while (*p >= 0) --p

#define ptr_for_binary(ks) (int*) (((ks)<<1)+1)
#define is_from_binary(p) ((unsigned int) p)&1
#define key_from_binary(p) (((int) p)>>1)

/* Macros for nclauses */
#define value_of(cl) -cl[1] 
#define cvalue(cl) cl[2]
#define dec_cvalue(cl) --cl[2]
#define set_ncl_val(cl,tt) cl[2]=tt
#define is_quasi_cl(cl) (cl[1] < 0) 

/***************************************************
 *
 *  VARIABLE Declaration
 *
 ***************************************************/

int TAPE_BLOCK;
TAPE *Tape_space;            /* a large storage space */
TAPE Tape_limit;             /* limit of each tape */
int Input_tape;
int Lemma_tape;
int last_Delete_num=0;

unsigned long Deleted_cells, Half_total_cells;

/* Total memory for variables: 6x4 bytes */

/* age-clauses */

/* Ages is a collection of clauses parititioned by ages.
   Each partition is a list of pointers to clause tapes.
*/

#ifdef AGE
int ***Ages;

#define age_count_inc(ptrs) ((int) ptrs[0])++
#define age_count_dec(ptrs) --((int) ptrs[0])
#define age_count(ptrs) (int) *ptrs
#define age_max(ptrs) (int) ptrs[1]
#define age_set_count(ptrs,c) ptrs[0]=(int*)c
#define age_set_max(ptrs,c) (int) ptrs[1]=(int*)c
#define age_begin() 2
#define age_push_clause(ptrs,ptr) ptrs[age_count_inc(ptrs)] = ptr


/***************************************************
 *
 *  FUN Declaration
 *
 ***************************************************/

ages_init() 
{
  int i;
  Ages = (int ***) malloc((Max_atom) * (sizeof(int **)));
  for (i = 0; i < Max_atom; i++) Ages[i] = NULL;
}

ages_insert_clause (a, cl)
     int a, *cl;
{
  int **ptrs = Ages[a];
  if (ptrs == NULL) {
    ptrs = Ages[a] = (int **) malloc(sizeof(int *) << 2);
    age_set_max(ptrs, 4);
    age_set_count(ptrs, 3);
    ptrs[2] = cl;
  } else {
    if (age_count(ptrs) == age_max(ptrs)) {
      int **ptrs2 = (int **) malloc((age_max(ptrs)<<1) * (sizeof(int *)));
      Ages[a] = ptrs2;
      memory_count(sizeof(int *) * age_max(ptrs), "AGE");
      if (ptrs2 == NULL) { fprintf(stderr, "ages_insert_clause.\n"); exit(1); }
      for (a = age_max(ptrs)-1; a >= 0; a--) ptrs2[a] = ptrs[a];
      age_max(ptrs2) += age_max(ptrs);
      free(ptrs);
      ptrs = ptrs2;
    }
    age_push_clause(ptrs, cl);
  }
}

ages_delete_clauses (a)
     int a;
{
  int **ptrs = Ages[a];
  if (ptrs) {
    for (a = age_begin(); a < age_count(ptrs); a++) {
      ptrs[a][1] = ptrs[a][2] = 0; /* mark as deleted */
      Deleted_cells -= ptrs[a][0];
      Delete_num++;
    }
    age_set_count(ptrs,age_begin());
  }
}
#endif

void tape_init ()
{
  int i;

  Deleted_cells = 0;

  if (Max_atom <= 100) MAX_TAPE = 64;
  else if (Max_atom <= 1000) MAX_TAPE = 128;
  else if (Max_atom <= 10000) MAX_TAPE = 256;
  else if (Max_atom <= 100000) MAX_TAPE = 512;
  else MAX_TAPE = 1024;
  TAPE_BLOCK = MAX_TAPE << (GROW+3);

  Tape_space = (TAPE *) malloc((MAX_TAPE) * (sizeof(TAPE)));
  memory_count((sizeof(TAPE *)*MAX_TAPE), "Tape_space");
  Tape_limit = (TAPE) malloc((MAX_TAPE) * (sizeof(int)));
  memory_count((sizeof(TAPE)*MAX_TAPE), "Tape_limit");

  for (i = 0; i < MAX_TAPE; i++) {
    Tape_space[i] = NULL;
    Tape_limit[i] = 0; 
  }
  Tape_space[0] = (TAPE) malloc((TAPE_BLOCK) * (sizeof(int)));
  if (!Tape_space[0]) { fprintf(stderr, "out of memory\n"); exit(0); }
  memory_count((sizeof(TAPE) * TAPE_BLOCK), "Tape");
  Input_tape = 0;
  Lemma_tape = MAX_TAPE-1;
  if (SELECT <= 3) { /* Tape[0] for Non-horn clauses */
    Input_tape = 1;
    Tape_space[1] = (TAPE) malloc((TAPE_BLOCK) * (sizeof(int)));
    if (!Tape_space[1]) { fprintf(stderr, "Out of memory\n"); exit(0); }
    memory_count((sizeof(TAPE) * TAPE_BLOCK), "Tape");
  }
  Half_total_cells = (TAPE_BLOCK*MAX_TAPE)>>1;
}

adjust_max_tape ()
{
  int i0 = Input_tape+((SELECT < 4)? 1 : 2);
  int i;
  
  i = (i0*(GROW+1)) << 2; 
  if (i < MAX_TAPE) {
    trace(1, printf("c MAX_TAPE is set to %d (%d, %d)\n", 
		    i, MAX_TAPE, i0));
    MAX_TAPE = i;
    Lemma_tape = MAX_TAPE-1;
  } else {
    trace(1, printf("c MAX_TAPE (%d) is not as big as %d\n", MAX_TAPE, i));
  }
}

int *get_tape (s, i)
     int s, i;
{
  if (Tape_space[i] == NULL) {
    if ((Tape_space[i] = (TAPE) malloc((TAPE_BLOCK) * (sizeof(int)))) == NULL) {
      fprintf(stderr, "ABEND, malloc returns NULL (out of memory).\n");
      fprintf(stderr, "%10.2f K has been mallocated.\n", memory_count_get());
      return NULL;
    }
    memory_count((sizeof(TAPE) * TAPE_BLOCK), "tape");
    /* printf("tape[%d] is ready (%d)\n", i, Input_tape);*/
  }
  Tape_limit[i] = s;
  return Tape_space[i];
}

int *get_tape_head0 (s, t)
     int s, t;
{
  int i, *p;

  if (Tape_limit[t]+s < TAPE_BLOCK) {
    p = Tape_space[t]+Tape_limit[t];
    Tape_limit[t] += s;
    return p;
  }

  if (Lemma_tape <= Input_tape+1) return NULL;
  t = --Lemma_tape;
  Tape_limit[t] = s;
  return Tape_space[t];
}

int *get_tape_head (s, t)
     int s, t;
{
  int i, *p;
  s += lit0();

  if (!Tape_space[t]) return get_tape(s, t);

  if (Tape_limit[t]+s < TAPE_BLOCK) {
    p = Tape_space[t]+Tape_limit[t];
    Tape_limit[t] += s;
    return p;
  }
  if (t == 0) { 
    t = Input_tape;
    if (Tape_limit[t]+s < TAPE_BLOCK) {
      p = Tape_space[t]+Tape_limit[t];
      Tape_limit[t] += s;
      return p;
    }
  }

  if (Lemma_tape <= Input_tape+1) {
    if (last_Delete_num < Delete_num) {
      clean_tapes();
      return get_tape_head0(s, Lemma_tape);
    } else {
      return NULL;
    }
  }

  if (t == Input_tape) t = ++Input_tape;
  else t = --Lemma_tape;
  return get_tape(s, t);
}

delete_clauses_on_tape ()
{
  int i, j, k, len, *cl, *cl2;

  trace(3, printf("Begin deleting clauses ...\n"));

  i = MAX_TAPE-1;
  while (i >= Lemma_tape) {
    cl = Tape_space[i];
    cl2 = cl+Tape_limit[i];
    while (cl < cl2) {
      len = size_of_clause(cl);
      if (undeleted_clause(cl) && len > DLENGTH+2 &&
	  L5[j=get_head(cl)]->value != FF &&
	  L5[k=get_tail(cl)]->value != FF) {
	set_lit_clauses_deleted(j);
	set_lit_clauses_deleted(k);
	Deleted_cells += len;
	Delete_num++;
	cl[1] = cl[2] = 0; /* mark as deleted */
      }
      cl += len;
    }
    i--;
  }

  /* clean the tapes? */
  if (Deleted_cells >= Half_total_cells) {
    clean_tapes();
  } else {
    if (DLENGTH > 5 && (Lemma_tape<<1) < (MAX_TAPE-Input_tape))
      DLENGTH--;

    /* we need to remove hts of deleted clauses */
    /* this operation can be done lazily */
    for (k = get_lit(Max_atom,1); k >= 2; k--) 
      if (lit_clauses_deleted(k)) {
	int **pts = L5[k]->hts;
	reset_lit_clauses_deleted(k);
	j = hts_count(pts);
	i = hts_begin(); 
	while (i < j) {
	  if (undeleted_head_tail(pts[i])) i++;
	  else pts[i] = pts[--j];
	}
	hts_set_count(pts,i);
      }
  }
}

clean_tapes ()
{
  int ck, i, j, j2, k, t2, *cl, *cl2, *tape, *tape2;
  int **ptrs; 

  i = t2 = MAX_TAPE-1;
  tape2 = Tape_space[t2];
  j2 = 0;

  while (i >= Lemma_tape) {
    int limit = Tape_limit[i];
    tape = Tape_space[i--];
    for (j=0; j < limit; ++j){ 
      if ((k=tape[j]) < 0) {
	/* beginning of clause */
	if (tape[j+1] == 0) {
	  j = j-k-1;  /* ++j later */
	} else {
	  if (j2-k >= TAPE_BLOCK) {
	    Tape_limit[t2] = j2;
	    tape2 = Tape_space[--t2];
	    tape2[0] = k;
	    j2 = 1;
	    cl = &tape[j];
	    cl2 = tape2;
	  } else {
	    cl = &tape[j];
	    cl2 = &tape2[j2];
	    tape2[j2++] = k;
	  }

	  /* cl, cl2 are needed to update gmark */
	  if (L5[get_head(cl)]->value == FF) {
	    ck = lit_var(get_tail(cl));
	    if (var_gmark(ck) == (int) cl) var_gmark(ck) = (int) cl2;
	  } else if (L5[get_tail(cl)]->value == FF) {
	    ck = lit_var(get_head(cl));
	    if (var_gmark(ck) == (int) cl) var_gmark(ck) = (int) cl2;
	  }
	}
      } else {
	tape2[j2++] = k;
      }
    }
  }
  Tape_limit[t2] = j2;
  for (j = Lemma_tape; j < t2; j++) Tape_limit[j] = 0;
  trace(1, printf("c %ld clauses deleted (total=%d), avg len = %d, %d tapes released\n",
		  Delete_num-last_Delete_num, Delete_num,
		  Deleted_cells/(Delete_num-last_Delete_num), 
		  t2-Lemma_tape));
  Lemma_tape = t2;
  Deleted_cells = 0;
  last_Delete_num = Delete_num;

  /* clean up the head_tail pointers from variables */
  for (i= get_lit(Max_atom,1); i >= 2; i--) {
    hts_set_count(L5[i]->hts,hts_begin());
  }

#ifdef AGE
  /* clean up the age pointers from Ages */
  for (i=1; i <= Level; i++) if (Ages[i]) {
    age_set_count(Ages[i],age_begin());
  }
#endif

  /* re-insert the head_tail_pointers */
  /* how to avoid repeating this work for input clauses? */
  i = 0; 
  while (i < MAX_TAPE) {
    cl = Tape_space[i];
    cl2 = cl+Tape_limit[i];
    while (cl < cl2) {
      if (is_normal(cl)) {
	/* insert head */
	ptrs = L5[get_head(cl)]->hts;
	push_head_tail(ptrs, cl+1);
	/* insert tail */
	ptrs = L5[get_tail(cl)]->hts;
	push_head_tail(ptrs, cl+2);

#ifdef AGE
	/* insert age */
	if (age_of_cl(cl)) ages_insert_clause(age_of_cl(cl), cl);
#endif
      }

      cl += size_of_clause(cl);
    }
    if (i == Input_tape) i = Lemma_tape;
    else i++;
  }
}

void print_clause (cl)
     TAPE cl;
{
  int i, j, k, v;

  if (is_from_binary(cl)) {
    i = key_from_binary(cl);
    printf("from binary %s%d\n", (i&1)? "" : "-", i>>1);
    return;
  }
  j=size_of_clause(cl);
  if (is_normal(cl)) {
    printf("size=%d ( ", j-lit0()); 
  } else {
    printf("size=%d cval=%d ( ", j-lit0(), cl[2]);
  }

  for (i = lit0(); i < j; i++) {
    v = cl[i];
    k = lit_var(v);
    if (is_head_tail(cl,i)) printf("^");
#ifdef DYNAMIC
    if (var_value_flexible(k)) { printf("?"); } 
#endif
    if (lit_sign(v)) printf("-");
    printf("%d%s ", k, (var_value(k) == TT)? "+" : ((var_value(k) == FF)? "-" : "")); 
  }
  if (is_normal(cl)) 
    printf(") h=%d t=%d a=%d\n", 
	   cl[1]+1-lit0(), cl[2]+1-lit0(), age_of_cl(cl));
  else printf(")\n");
}

void print_clauses_on_tape (tape1, tape2)
     int tape1, tape2;
{
  int i;

  printf("all clauses on tapes %d-%d\n", tape1, tape2);
  for (i = tape1; i <= tape2; i++) {
    int *cl2, *cl = Tape_space[i];
    cl2 = cl+Tape_limit[i];
    while (cl < cl2) {
      if (size_of_clause(cl) < 6) {
	break;
      } else { 
	print_clause(cl);
	cl += size_of_clause(cl);
      }
    }
  }
}

check_clause (cl)
     TAPE cl;
{
  int j;
  for (j = lit0(); j < size_of_clause(cl); j++) if (L5[cl[j]]->value == TT) return 0;
  trace(2, { printf("False cl: "); print_clause(cl);});
  return 1;
}

#ifdef QUASICLAUSE
check_nclause (cl)
     TAPE cl;
{
  int j, s=0;
  for (j = lit0(); j < size_of_clause(cl); j++) if (L5[cl[j]]->value) s++;
  s = s+cl[1];
  if (s > 0) { 
    trace(2, { printf("False cl: "); print_clause(cl); }); return 1;
  }
  return 0;
}
#endif

var5_check_model ()
{
  int i, j, *cl;

  trace(2, printf("c Checking model ...\n"));

  for (i = 0; i <= Input_tape; i++) {
    TAPE cl = Tape_space[i];
    int limit = Tape_limit[i];

    while (limit>0) {
       if (is_normal(cl)) {
	if (check_clause(cl)) return 1;
#ifdef QUASICLAUSE
      } else if (undeleted_clause(cl)) {
	if (check_nclause(cl)) { return 1; }
#endif
      }
      j = size_of_clause(cl);
      limit -= j;
      cl += j;
    }
  }
  return 0;
}

int *insert_clause_to_tape (cl_arr, total, plen, type)
     int cl_arr[], total, plen, type;
{
  int i, value=cl_arr[total], *cl, *p;

  if (type == NORM) { 
    if (total == 1) { 
      i = cl_arr[0];
      assign_value1(i);
#ifdef MINONE
      if (MINVAR && !(lit_sign(i)) && (++Current_best > MINVAR)) { 
	 var5_fail_end(NULL); 
      }
#endif
      return NULL; 
    }
    if (total == 2) { 
      insert_binary_clause(cl_arr[0], cl_arr[1]); 
      return NULL; 
    }
  } else if (type == GTEQ || type == GRTR) { 
    if (type == GRTR) value++;
    if (value == 0) return NULL; 
    if (value == 1) type = NORM; 
    else {
      /* changing GTEQ to LSEQ */
      for (i=0; i < total; i++) cl_arr[i] = NEG(cl_arr[i]);
      value = total-value;
      type = LSEQ;
    }
  } else if (type == LESS) { 
    type = LSEQ; value--; 
  } 

  if (Branch_open) {
    cl = get_tape_head(total, Lemma_tape);
  } else if (type) { /* type != NORM */
    NH_num++;
    POS_num++;
    cl = get_tape_head(total, 0);
  } else if (plen > 1) {
    NH_num++;
    if (plen == total) {
      POS_num++;
      cl = get_tape_head(total, 0);
    } else
      cl = get_tape_head(total, Input_tape);
  } else {   
    cl = get_tape_head(total, Input_tape);
  }

  if (cl == NULL) {
    fprintf(stderr, "No space for input clauses (because of -g%d)\n", GROW);
    exit(1);
  } 
  set_clause_size(cl, total);
  p = cl;

#ifdef QUASICLAUSE
  if (type) { 
    *(++p) = -value;  /* original value */
    *(++p) = value;   /* current value */
    ++p; /* skip a position for age */
    for (i = 0; i < total; i++) {
      int j;
      struct nlink *nl = get_nlink();
      set_lit(++p, cl_arr[i]);
      nl->cl = cl;
      j = (*p)^1;
      nl->next = L5[j]->ncls;
      L5[j]->ncls = nl;
      weight_add(*p,value); 
      weight_add(j,value); 
    }
  } else 
#endif
  { 
    /* set the head */
    set_head_tail(++p, lit0()); 
    insert_head_tail(p,cl_arr[0]);
    /* set the tail */
    set_head_tail(++p, total+lit0()-1); 
    insert_head_tail(p,cl_arr[total-1]);
    ++p; /* skip a position for age */
    for (i=0; i < total; i++) {
      set_lit(++p, cl_arr[i]);
      weight_add1(*p); 
    }
  }

  trace(9, { printf("IN(%d): ", type); print_clause(cl);});
  return cl;
}

int *insert_lemma_to_tape (cl_arr, total, p1, p2)
     int cl_arr[], total, p1, p2;
{
  int i, sig, *cl, *p, age = 0;
#ifdef AGE
  age = cl_arr[total];
#endif

  if (p1 > p2) { i = p1; p1 = p2; p2 = i; }
  if (total == 3) {
    if (p1 != 0) {
      i = cl_arr[0]; cl_arr[0] = cl_arr[p1]; cl_arr[p1] = i; p1 = 0;
    } else if (p2 != 2) {
      i = cl_arr[2]; cl_arr[2] = cl_arr[p2]; cl_arr[p2] = i; p2 = 2;
    }
  }

  cl = get_tape_head(total, (!age && (total<10))? Input_tape : Lemma_tape);
  if (cl == NULL) {
    printf("c no space for new clauses (-g%d)\n", GROW);
    GROW = 0;
    return NULL;
  }
  set_clause_size(cl, total);
#ifdef AGE
  set_age_of_cl(cl, age);
#endif
  p = cl;

  /* set the head */
  set_head_tail(++p, p1+lit0());
  insert_head_tail(p,cl_arr[p1]);

  /* set the tail */
  set_head_tail(++p, p2+lit0());
  insert_head_tail(p,cl_arr[p2]);

  ++p; /* skip a position for age */
  for (i=0; i < total; i++) {
    set_lit(++p, cl_arr[i]); 
    weight_add1(*p); 
  }

  trace(3, { printf("NEW(%d): ", Level); print_clause(cl);});
  return cl;
}
